/*
** Module   : DatabaseTriggerManager.java
** Abstract : Database trigger manager.
**
** Copyright (c) 2009-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------- Description ----------------------------
** 001 SVL 20091019   @44152 Created initial version. Supports assign and 
**                           delete schema triggers.
** 002 CA  20100527   @44879 Added support for CREATE and "no record is 
**                           available triggers". The CREATE, DELETE and 
**                           ASSIGN triggers will execute on undo too, but the
**                           passed DMO will not be the buffer proxy; instead,
**                           it will be the actual record instance.
** 003 SVL 20130213          Added (de)registerDatabaseTrigger and disable*Triggers functions.
** 004 OM  20131107          Rewritten entire class.
** 005 OM  20140107          Removed openReadOnlyScope, added error message for FIND attempts on
**                           read-only buffers created by WRITE events.
** 006 OM  20140711          Rewrote the initialization. The trigger list is read from xml file 
**                           instead of scanning for SchemaTrigger annotation.
*/

package com.goldencode.p2j.persist.trigger;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class handles the management of trigger system for a database. It maintains a stack of
 * scopes (a layer of triggers for each subroutine block).
 * One instance of this class is associated an user context and, because this class implements
 * <code>Scopeable</code> it is initialized in <code>StandardServer.invoke()</code>, before the
 * topmost "startup" scope is pushed.
 * The <code>DatabaseTriggerManager</code> is notified when the execution enters and leaves
 * subroutine blocks so it can do the management of trigger layers.
 */
public class DatabaseTriggerManager
implements Scopeable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(DatabaseTriggerManager.class);
   
   /** Context local map of instances of this class, mapped by theirs database name . */
   private static final ContextLocal<DatabaseTriggerManager> context =
         new ContextLocal<DatabaseTriggerManager>()
         {
            protected DatabaseTriggerManager initialValue()
            {
               return new DatabaseTriggerManager();
            }
         };
   
   /** The 'global' layer of schema triggers. */
   private static final TriggerBlockLayer schemaTriggers = new TriggerBlockLayer();
   
   /** The <code>trigger</code> element in schema-triggers.xml. */
   public static final String TRIGGER_ELEMENT = "trigger";
   
   /** The <code>class</code> attribute of trigger element in schema-triggers.xml. */
   public static final String CLASS_ATTRIBUTE = "class";
   
   /** The <code>buffer</code> attribute of trigger element in schema-triggers.xml. */
   public static final String BUFFER_ATTRIBUTE = "buffer";
   
   /** The <code>property</code> attribute of assign trigger element in schema-triggers.xml. */
   public static final String PROPERTY_ATTRIBUTE = "property";
   
   /**
    * The stack of all layer of triggers handled by this manager.
    */
   private Deque<TriggerBlockLayer> stack = new ArrayDeque<>();
   
   /**
    * Private constructor. An instance of this class can be obtained only using get() method.
    * Registers this object with TransactionManager as a Scopeable instance.
    */
   private DatabaseTriggerManager()
   {
      // I want to receive enter/leave bloc events
      TransactionManager.registerScopeable(this);
      
      if (LOG.isLoggable(Level.INFO))
      {
         LOG.log(Level.INFO, "DatabaseTriggerManager registered with TransactionManager.");
      }
   }
   
   /**
    * Initialization of <code>DatabaseTriggerManager</code> static internal data.
    * The <code>schemaTriggers</code> xml file is parsed and from each valid entry, the trigger
    * class is probed if class and annotations are correct. If successful, the class is loaded
    * for the declared buffer and even. Otherwise the entry is discarded.
    *
    * @param   schemaTriggers
    *          the location from where the schemaTriggers list is loaded (as declared in
    *          package statement)
    */
   @SuppressWarnings("unchecked")
   public static void initialize(String schemaTriggers)
   {
      schemaTriggers = schemaTriggers.replace('.', '/') + "schema-triggers.xml";
      InputStream xmlIs = Thread.currentThread().getContextClassLoader().getResourceAsStream(
            schemaTriggers);
      Document dom = null;
      try
      {
         dom = XmlHelper.parse(xmlIs);
      }
      catch (ParserConfigurationException | SAXException | IOException e)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Failed to parse schema trigger list: " + schemaTriggers);
         }
         return;
      }
      
      Element root = dom.getDocumentElement();
      if (root == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Incorrect format for schema trigger list: " + schemaTriggers);
         }
         return; // no schemaTriggers ?
      }
      
      NodeList trElements = root.getElementsByTagName(TRIGGER_ELEMENT);
      if (trElements == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Empty list of triggers : " + schemaTriggers);
         }
         return; // no schemaTriggers ?
      }
      
      int registeredTriggers = 0;
      final int length = trElements.getLength();
      for (int i = 0; i < length; i++)
      {
         Element triggerDef = (Element) trElements.item(i);
         
         String triggerCandidate = triggerDef.getAttribute(CLASS_ATTRIBUTE);
         
         if (triggerCandidate.isEmpty())
         {
            // the trigger procedure could not be found for this schema trigger, skip it
            continue;
         }
         
         Class<?> ufo = null;
         try
         {
            ufo = Class.forName(triggerCandidate);
         }
         catch (ClassNotFoundException e)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING, "Failed to load schema trigger " + triggerCandidate);
            }
            continue;
         }
         if (!DatabaseTrigger.class.isAssignableFrom(ufo) ||
               !ufo.isAnnotationPresent(SchemaTrigger.class))
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING, "Not a schema trigger: " + triggerCandidate);
            }
            continue;
         }
         
         // not an ufo any more
         Class<? extends DatabaseTrigger> trigger = (Class<? extends DatabaseTrigger>) ufo;
         
         String property = triggerDef.getAttribute(PROPERTY_ATTRIBUTE);
         String buffer = triggerDef.getAttribute(BUFFER_ATTRIBUTE);
         SchemaTrigger annotation = trigger.getAnnotation(SchemaTrigger.class);
         Class<? extends Buffer> bufferCls = annotation.buffer(); // buffer
         DatabaseEventType det = annotation.event();
         String assignProperty = annotation.field(); // legacy of property
         boolean override = annotation.overrideable();
         boolean hasOldBuffer = annotation.hasOldBuffer();
         
         DatabaseTriggerManager.schemaTriggers.setTrigger(trigger,
                                                          det,
                                                          bufferCls,
                                                          property,
                                                          null, /* no container needed */
                                                          override,
                                                          hasOldBuffer);
         ++ registeredTriggers;
      }
      
      if (LOG.isLoggable(Level.INFO))
      {
         LOG.log(Level.INFO, "Loaded " + registeredTriggers + " schema schemaTriggers.");
      }
   }
   
   /**
    * Obtain the context-local DatabaseTriggerManager object.
    *
    * @return  the DatabaseTriggerManager for current context.
    */
   public static DatabaseTriggerManager get()
   {
      return context.get();
   }

   /**
    * Register a session database trigger within the current scope. Any previously
    * registered session trigger on current block is silently removed. This overloaded method is
    * always not overriding the schema trigger and is not declaring an <code>OLD [BUFFER]</code>.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, false, false);
   }

   /**
    * Register a session database trigger within the current scope. Any previously
    * registered session trigger on current block is silently removed. This overloaded method
    * not declaring an <code>OLD [BUFFER]</code>.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    * @param    override
    *           <code>true</code> if OVERRIDE keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, override, false);
   }

   /**
    * Register a session database trigger within the current scope. Any previously
    * registered session trigger on current block is silently removed.
    *
    * @param   event
    *          Database event type.
    * @param   bufferCls
    *          The class of the buffer to which the event is applied.
    * @param   triggerCls
    *          The class of the trigger.
    * @param   container
    *          The containing object (all triggers are implemented as non-static inner classes,
    *          so they must have a reference to the containing class).
    * @param   override
    *          <code>true</code> if OVERRIDE keyword is specified.
    * @param   hasOldBuffer
    *          <code>true</code> if an OLD [BUFFER] keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override,
                                              boolean hasOldBuffer)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      // check if this event is overrideable
      if (override && !isSchemaTriggerOverrideable(event, bufferCls))
      {
         ErrorManager.recordOrThrowError(3226,
            "Attempt to OVERRIDE a schema " + event.toString().toLowerCase() + " trigger");
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.setTrigger(triggerCls, event, bufferCls, null, container, override, hasOldBuffer);
   }

   /**
    * Register an <code> ASSIGN</code> session database trigger within the current scope. Any
    * previously registered session trigger on current block is silently removed. This overloaded
    * method is always not overriding the schema trigger and is not declaring an
    * <code>OLD [VALUE]</code>.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    */
   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container)
   {
      registerDatabaseTrigger(event, bufferCls, property, triggerCls, container, false, false);
   }

   /**
    * Register an <code>ASSIGN</code> session database trigger within the current scope. Any
    * previously registered session trigger on current block is silently removed. This overloaded
    * method is not declaring an <code>OLD [VALUE]</code>.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    * @param    override
    *           <code>true</code> if <code>OVERRIDE</code> keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override)
   {
      registerDatabaseTrigger(event, bufferCls, property, triggerCls, container, override, false);
   }

   /**
    * Register an <code>ASSIGN</code> session database trigger within the current scope. Any previously
    * registered session trigger on current block is silently removed.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as non-static inner classes,
    *           so they must have a reference to the containing class).
    * @param    override
    *           <code>true</code> if <code>OVERRIDE</code> keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event, // always ASSIGN
                                              Class<? extends Buffer> bufferCls,
                                              String property,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override,
                                              boolean hasOldValue)
   {
      // the first parameter is always DatabaseEventType.ASSIGN so it can be dropped from the
      // list of parameters for this kind of events, but we keep it to have consistency
      // with invocation for other events.
      if (event != DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Use this form of register only for ASSIGN triggers.");
      }
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      // check if this event is overrideable
      if (override && !isSchemaTriggerOverrideable(DatabaseEventType.ASSIGN, bufferCls))
      {
         ErrorManager.recordOrThrowError(3226, "Attempt to OVERRIDE a schema assign trigger");

         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.setTrigger(triggerCls, event, bufferCls, property, container, override, hasOldValue);
   }

   /**
    * Unregister a session database trigger for the given event type and buffer class.
    * This is only called when a <code>ON Assign OF &lt;table.field&gt; REVERT</code>.
    *
    * @param   event
    *          Database event type.
    * @param   bufferCls
    *          The class of the buffer to which the event is applied.
    * @param   property
    *          The property that is deregistered from <code>ASSIGN</code> trigger.
    */
   public static void deregisterDatabaseTrigger(DatabaseEventType event,
                                                Class<? extends Buffer> bufferCls,
                                                String property)
   {
      // the first parameter is always DatabaseEventType.ASSIGN so it can be dropped from the
      // list of parameters for this kind of events, but we keep it to have consistency
      // with invocation for other events.

      if (event != DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Use this form of register only for ASSIGN triggers.");
      }

      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.removeTrigger(event, bufferCls, property);
   }

   /**
    * Unregister a session database trigger for the given event type and buffer class.
    * This is only called when a <code>ON &lt;event&gt; OF &lt;table&gt; REVERT</code>.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    */
   public static void deregisterDatabaseTrigger(DatabaseEventType event,
                                                Class<? extends Buffer> bufferCls)
   {
      // the first parameter is always DatabaseEventType.ASSIGN so it can be dropped from the
      // list of parameters for this kind of events, but we keep it to have consistency
      // with invocation for other events.

      if (event == DatabaseEventType.ASSIGN)
      {
         throw new IllegalArgumentException("Not executing ASSIGN with this method.");
      }

      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         // sanity check failed
         return;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.removeTrigger(event, bufferCls, null);
   }

   /**
    * Disables dump triggers (<code>FIND</code> events) for the current block and any block of a
    * callee from this block (function, internal/external procedure).
    *
    * @param   bufferCls
    *          The buffer for which to disable dump triggers.
    *
    * @return  <code>true</code> if the filter was successfully applied. If <code>false</code>
    *          is returned then the internal stack is broken because of bad block management.
    */
   public static boolean disableDumpTriggers(Class<? extends Buffer> bufferCls)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         return false;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.disableTriggers(bufferCls, true, false, false);
      return true;
   }

   /**
    * Disables load triggers (<code>CREATE</code>, <code>ASSIGN</code>, <code>WRITE</code> events)
    * for the current block. The replication triggers are also disabled.
    *
    * @param   bufferCls
    *          The buffer for which to disable load triggers.
    *
    * @return  <code>true</code> if the filter was successfully applied. If <code>false</code>
    *          is returned then the internal stack is broken because of bad block management.
    */
   public static boolean disableLoadTriggers(Class<? extends Buffer> bufferCls)
   {
      return disableLoadTriggers(bufferCls, false);
   }

   /**
    * Disables load triggers (<code>CREATE</code>, <code>ASSIGN</code>, <code>WRITE</code> events)
    * for the current block.
    *
    * @param   bufferCls
    *          The buffer for which to disable load triggers.
    * @param   allowReplication
    *          If <code>true</code> the replication triggers will not be disabled for this block.
    *
    * @return  <code>true</code> if the filter was successfully applied. If <code>false</code>
    *          is returned then the internal stack is broken because of bad block management.
    */
   public static boolean disableLoadTriggers(Class<? extends Buffer> bufferCls,
                                             boolean allowReplication)
   {
      DatabaseTriggerManager manager = get();
      if (manager.stack.isEmpty())
      {
         return false;
      }

      TriggerBlockLayer tbl = manager.stack.peekFirst();
      tbl.disableTriggers(bufferCls, false, true, allowReplication);

      return true;
   }

   /**
    * Executes trigger suite for a base event. (Except for <code>ASSIGN</code> triggers.)
    * The suite includes the session, simple schema and replication triggers that have been
    * defined/assigned for the <code>det</code> base event. The order in which they execute
    * depends on <code>det</code>: session triggers are always executed first and replication
    * triggers last, except for the <code>FIND</code> triggers, where schema triggers execute
    * execute before session triggers and there is no replication trigger to execute.
    *
    * @param   det
    *          The trigger type to be executed.
    * @param   bufferCls
    *          The buffer class for the dmo that fired the trigger.
    * @param   newVal
    *          The buffer that fired the trigger.
    * @param   oldDmoValue
    *          When the (WRITE) trigger declares an 'old' buffer value, this is passed to
    *          trigger procedure as r/o buffer.
    */
   public static void trigger(DatabaseEventType det,
                              Class<? extends Buffer> bufferCls,
                              Buffer newVal,
                              Persistable oldDmoValue)
   {
      // TODO: Persistence.load():2029: Object o = query.uniqueResult();
      // TODO: can trigger an uncatched WRITE database event

      DatabaseTriggerManager manager = get();
      TriggerBlockLayer.EnableStatus triggerStatus = manager.getTriggerStatus(det, bufferCls);
      if (triggerStatus == TriggerBlockLayer.EnableStatus.DISABLED)
      {
         // triggers are disabled for this kind of event / table
         return;
      }

      if (triggerStatus == TriggerBlockLayer.EnableStatus.ENABLED)
      {
         // look for the triggers for this event / bufferCls
         TriggerData sessionTrigger = manager.peekTrigger(det, bufferCls, null);
         TriggerData schemaTrigger = schemaTriggers.getTrigger(det, bufferCls, null);

         boolean overrideSchemaTriggers = sessionTrigger != null && sessionTrigger.isOverride();

         // except for the FIND case, the session trigger executes first (if it ever does)
         if (sessionTrigger != null && det != DatabaseEventType.FIND)
         {
            // each trigger invocation is done using a different instance of the OLD, r/o buffer
            boolean sessOldBuffer = oldDmoValue != null && sessionTrigger.hasOldBuffer();
            Buffer oldBuffer = sessOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;
            fireTrigger(det, sessionTrigger, newVal, oldBuffer);
         }

         if (schemaTrigger != null && !overrideSchemaTriggers)
         {
            // each trigger invocation is done using a different instance of the OLD, r/o buffer
            boolean schOldBuffer = oldDmoValue != null && schemaTrigger.hasOldBuffer();
            Buffer oldBuffer = schOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;
            fireTrigger(det, schemaTrigger, newVal, oldBuffer);
         }

         // in the FIND case, the session trigger executes second (if it ever does)
         if (sessionTrigger != null && det == DatabaseEventType.FIND)
         {
            // FIND triggers do not have an old buffer
            fireTrigger(det, sessionTrigger, newVal, null);
         }
      }

      // replication triggers ignores the overrideable attribute, they always trigger
      TriggerData replicationTrigger = schemaTriggers.getReplicationTrigger(det, bufferCls);
      det = getReplicationEvent(det);
      if (replicationTrigger != null && det != null)
      {
         boolean replOldBuffer = oldDmoValue != null && replicationTrigger.hasOldBuffer();
         Buffer oldBuffer = replOldBuffer ? createWriteOldBuffer(bufferCls, oldDmoValue) : null;

         fireTrigger(det, replicationTrigger, newVal, oldBuffer);
      }
   }

   /**
    * Convert a base database event into the replication event. If they are already replication
    * events, they are returned. If they do not have a replication event associated,
    * <code>null</code> is returned.
    *
    * @param   det
    *          The base trigger event.
    *
    * @return  The replication event of <code>det</code>. For FIND and ASSIGN <code>null</code>
    *          is returned.
    */
   public static DatabaseEventType getReplicationEvent(DatabaseEventType det)
   {
      switch (det)
      {
         case CREATE:
         case REPLICATION_CREATE:
            return DatabaseEventType.REPLICATION_CREATE;
         case DELETE:
         case REPLICATION_DELETE:
            return DatabaseEventType.REPLICATION_DELETE;
         case WRITE:
         case REPLICATION_WRITE:
            return DatabaseEventType.REPLICATION_WRITE;
         default:
            return null;
      }
   }

   /**
    * Executes an ASSIGN trigger.
    *
    * @param   bufferCls
    *          The buffer class for the dmo that fired the trigger.
    * @param   newVal
    *          The buffer that contains the field that fired the trigger.
    * @param   oldValue
    *          When the ASSIGN trigger declares an 'old' value, this is passed to trigger
    *          procedure. The procedure code is allowed to modify this value without generating
    *          any error, however, they are lost as only a duplicate is passed on.
    */
   public static void triggerAssign(Class<? extends Buffer> bufferCls,
                                    Buffer newVal,
                                    String property,
                                    BaseDataType oldValue)
   {
      DatabaseTriggerManager manager = get();
      if (manager.getTriggerStatus(DatabaseEventType.ASSIGN, bufferCls) !=
               TriggerBlockLayer.EnableStatus.ENABLED)
      {
         // triggers are disabled for this kind of event / buffer
         // there are no replication triggers for ASSIGN event
         return;
      }

      // look for the trigger
      TriggerData sessionTrigger = manager.peekTrigger(DatabaseEventType.ASSIGN,
                                                       bufferCls,
                                                       property);

      //execute session trigger first
      if (sessionTrigger != null)
      {
         fireAssignTrigger(sessionTrigger, newVal, oldValue.duplicate());

         if (sessionTrigger.isOverride())
         {
            // if session trigger overrides then we return now without executing schema trigger
            return;
         }
      }

      // the schema trigger executes second
      TriggerData schemaTrigger = schemaTriggers.getTrigger(DatabaseEventType.ASSIGN,
                                                            bufferCls,
                                                            property);
      if (schemaTrigger != null)
      {
         fireAssignTrigger(schemaTrigger, newVal, oldValue.duplicate());
         // if trigger object failed an exception have already been printed to console
      }
   }

   /**
    * Provides a notification that a new scope is about to be entered. The manager only acts on
    * top-level scope. Each time a new scope is entered a fresh new layer of triggers is added
    * to internal stack.
    */
   @Override
   public void scopeStart()
   {
      // triggers get notifications at all top level blocks
      if (TransactionManager.isTopLevelBlock())
      {
         stack.addFirst(new TriggerBlockLayer()); // push();
      }
   }

   /**
    * Provides a notification that an external scope is about to be deleted.
    */
   @Override
   public void scopeDeleted()
   {
      // no op
   }

   /**
    * Provides a notification that a scope is about to be exited. The manager only acts on
    * top-level scope. Each time a scope is exited a the last layer of triggers is removed from
    * internal stack.
    */
   @Override
   public void scopeFinished()
   {
      // triggers get notifications at all top level blocks
      if (TransactionManager.isTopLevelBlock())
      {
         if (stack.isEmpty())
         {
            LOG.log(Level.WARNING, "Invalid scope finished.");
         }
         else
         {
            stack.removeFirst(); // pop();
         }
      }
   }

   /**
    * Tests if a schema trigger is overrideable. In case the trigger is declared as overideable,
    * <code>true</code> is returned. Otherwise, 3223 code is <code>recordOrThrowError</code>.
    * If no error is acting <code>false</code> will be returned.
    *
    * @param   event
    *          The database event of the trigger to be tested.
    * @param   bufferCls
    *          The class buf for the table the trigger is defined.
    *
    * @return  <code>true</code> if respective trigger can be overrided by a session trigger.
    */
   private static boolean isSchemaTriggerOverrideable(DatabaseEventType event,
                                                      Class<? extends Buffer> bufferCls)
   {
      // check if this event is overrideable
      TriggerData schemaTrigger = schemaTriggers.getTrigger(event, bufferCls, null);
      if (schemaTrigger != null && !schemaTrigger.isOverride())
      {
         // detect table name
         String tableName = "?";
         for (Class<?> next : bufferCls.getInterfaces())
         {
            if (DataModelObject.class.isAssignableFrom(next))
            {
               tableName = next.getSimpleName();
               break;
            }
         }

         String msg = String.format("Attempt to OVERRIDE a schema %s trigger for %s",
                                    event,
                                    tableName);
         // this is a non overrideable schema trigger:
         ErrorManager.recordOrThrowError(3223, msg);
         return false;
      }
      return true;
   }

   /**
    * Create a trigger object instance based on information from a <code>TriggerData</code>
    * container.
    *
    * @param   triggerData
    *          The container that holds information about the trigger to be built.
    *
    * @return  An instance of the trigger as configured in <code>triggerData</code>.
    */
   @SuppressWarnings("unchecked")
   private static DatabaseTrigger<Buffer> instantiateTrigger(TriggerData triggerData) {
      try
      {
         if (triggerData.getTriggerContainer() == null)
         {
            // if the container is null then this is (probably) a schema trigger
            // being a top-level public class, it does not need a container for constructing
            return triggerData.getTriggerClass().newInstance();
         }

         Class[] ctorParmlist  = new Class[] { triggerData.getTriggerContainer().getClass() };
         Constructor<? extends DatabaseTrigger> ctor =
               triggerData.getTriggerClass().getConstructor(ctorParmlist);
         return ctor.newInstance(triggerData.getTriggerContainer());
      }
      catch (InstantiationException | IllegalAccessException |
            NoSuchMethodException | InvocationTargetException e)
      {
         e.printStackTrace();
         // could not instantiate the trigger class
         return null;
      }
   }

   /**
    * Extract only the DMO schema-specific interface (i.e., exclude the Buffer interface)
    *
    * @param   bufferCls
    *          Interface which defines DMO's public API plus all {@link Buffer} methods.
    * @param   oldDmoValue
    *          The read-only value.
    *
    * @return  A temporary read-only buffer that can be used as the OLD 'parameter' in a
    *          WRITE trigger.
    */
   private static Buffer createWriteOldBuffer(Class<? extends Buffer> bufferCls,
                                              Persistable oldDmoValue)
   {
      Class<?> dmoIface = null;
      Class<?>[] superIfaces = bufferCls.getInterfaces();
      for (Class<?> next : superIfaces)
      {
         if (DataModelObject.class.isAssignableFrom(next))
         {
            dmoIface = next;
            break;
         }
      }

      if (dmoIface == null)
      {
         LOG.log(Level.WARNING, "Could not create old buffer for WRITE trigger.");
         return null;
      }

      Buffer oldValue = RecordBuffer.defineReadOnly(bufferCls,
                                                    DatabaseManager.getSchemaByInterface(dmoIface),
                                                    oldDmoValue);
      return oldValue;
   }

   /**
    * Instantiates an object of the ASSIGN trigger and fires it.
    *
    * @param   triggerData
    *          Information about how the trigger will be built.
    * @param   newVal
    *          The buffer that contains the field that fired the trigger.
    * @param   oldValue
    *          The old value of the field. May be null if the trigger did not declared it.
    */
   private static void fireAssignTrigger(TriggerData triggerData,
                                         Buffer newVal,
                                         BaseDataType oldValue)
   {
      final DatabaseTrigger<Buffer> trigger = instantiateTrigger(triggerData);
      if (trigger != null)
      {
         // TODO: for "ASSIGN/NEW" TRIGGER PROCEDURE: they do not have a Buf reference
         // * new annotation must be added at generation
         // * at this moment (in this method) the annotation should be checked
         // * if present, use the Undoable/BaseDataType signature method using the reference to
         //       the field from trigger's annotation or a copy of the actual value
         // * the trigger is allowed to change it, if the case, the new value must be set back
         //       to respective property of newVal.
         if (triggerData.hasOldBuffer())
         {
            trigger.assign(newVal, oldValue);
         }
         else
         {
            trigger.assign(newVal);
         }
      }
      // if trigger object failed an exception have already been printed to console
   }

   /**
    * Instantiates a trigger object associated with the specified database event type and
    * fires it.
    *
    * @param   det
    *          The database event type to be fired.
    * @param   triggerData
    *          Information about how the trigger will be built.
    * @param   buffer
    *          The buffer that fired the trigger.
    * @param   oldValue
    *          The old value of the buffer. May be null if the trigger did not declared it.
    */
   private static void fireTrigger(DatabaseEventType det,
                                   TriggerData triggerData,
                                   Buffer buffer,
                                   Buffer oldValue)
   {
      final DatabaseTrigger<Buffer> trigger = instantiateTrigger(triggerData);
      if (trigger == null)
      {
         // if trigger object failed an exception have already been printed to console
         return;
      }

      switch (det)
      {
         case ASSIGN:
            throw new IllegalArgumentException("Not executing ASSIGN with this method.");
            // break;
         case CREATE:
            trigger.create(buffer);
            break;
         case DELETE:
            trigger.delete(buffer);
            break;
         case FIND:
            trigger.find(buffer);
            break;
         case WRITE:
            if (triggerData.hasOldBuffer() /* && oldValue != null ? */)
            {
               trigger.write(buffer, oldValue);
            }
            else
            {
               trigger.write(buffer);
            }
            break;
         case REPLICATION_CREATE:
            trigger.replicationCreate(buffer);
            break;
         case REPLICATION_DELETE:
            trigger.replicationDelete(buffer);
            break;
         case REPLICATION_WRITE:
            if (triggerData.hasOldBuffer() /* && oldValue != null ? */)
            {
               trigger.replicationWrite(buffer, oldValue);
            }
            else
            {
               trigger.replicationWrite(buffer);
            }
            break;
      }
   }

   /**
    * Handle trigger- related errors. The error is displayed in status bar then the execution of
    * the procedure is abruptly interrupted because all trigger errors are fatal errors.
    *
    * @param   errId
    *          The id of the error.
    * @param   det
    *          The database event that caused the error.
    * @param   tableOrBufferName
    *          The name of the table (or buffer in some cases) that caused the error
    * @param   fieldName
    *          In the case of <code>ASSIGN</code> trigger the field that caused the error.
    */
   public static void handleError(int errId,
                                  DatabaseEventType det,
                                  String tableOrBufferName,
                                  String fieldName)
   {
      String msg = null;
      boolean displayOnly = false;

      switch (errId)
      {
         case 3168:
            msg = "Attempt to CREATE a " + tableOrBufferName + "record in " + det + "-TRIGGER " +
                  ProcedureManager.lastProcedure() + " while a trigger is executing";
            break;

         case 3169:
            msg = "Attempt to DELETE a " + tableOrBufferName + "record in " + det + "-TRIGGER " +
                  ProcedureManager.lastProcedure() + " while a trigger is executing";
            break;

         case 2873:
            assert (det == DatabaseEventType.FIND);
            msg = "Cannot change cursor of buffer " + tableOrBufferName +
                  "when executing a trigger";
            displayOnly = true;
            break;

         case 2872:
            assert (det == DatabaseEventType.WRITE);
            msg = "Attempt to FIND old trigger buffer during WRITE trigger for " +
                  tableOrBufferName;
            break;

         case 1003:
            assert (det == DatabaseEventType.ASSIGN);
            msg = "Cycle in procedure calls detected in 'ASSIGN-TRIGGER " +
                  ProcedureManager.lastProcedure() + "' while executing ASSIGN trigger for " +
                  tableOrBufferName + "." + fieldName;
            break;

         case 2665:
            // this is handled in RecordBuffer
            msg = "Attempt to update a Read-Only buffer for table " + tableOrBufferName;
            break;

         case 3129:
            // TODO: I found this error when (assign) trigger is fired after buffer RELEASE (if this is possible ?)
            assert (det == DatabaseEventType.ASSIGN);
            int[] errCodes = {91, 3129};
            String[] errMsgs = {"** No " + tableOrBufferName + " record is available",
                                "Could not extract value from " + tableOrBufferName +
                                      " (rpos 5) for an ASSIGN trigger" };
            ErrorManager.recordOrThrowError(errCodes, errMsgs, false, true);
            return;
      }

      if (displayOnly)
      {
         ErrorManager.displayError(errId, msg);
      }
      else
      {
         ErrorManager.showErrorAndAbend(errId, msg);
      }
   }

   /**
    * Looks into the stack of layers and check if a trigger is enabled for a database event.
    * If the trigger was not disabled at all, <code>EnableStatus.ENABLED</code> is returned. If
    * the trigger was disabled, the <code>EnableStatus.DISABLED</code> is returned unless the
    * events has a replication pair and it was not disable using <code>ALLOW-REPLICATION</code>,
    * when <code>EnableStatus.REPLICATION_ONLY</code> is returned.
    *
    * @param   det
    *          The database (base) event.
    * @param   bufferCls
    *          The buffer.
    *
    * @return  the status of the trigger on the buffer.
    */
   private TriggerBlockLayer.EnableStatus getTriggerStatus(DatabaseEventType det,
                                                           Class<? extends Buffer> bufferCls)
   {
      for (TriggerBlockLayer tbl : stack)
      {
         TriggerBlockLayer.EnableStatus status = tbl.getTriggerStatus(det, bufferCls);
         if (status != TriggerBlockLayer.EnableStatus.UNKNOWN)
         {
            return status;
         }
      }

      // if no disable-triggers were found in the whole stack then the trigger is enabled
      return TriggerBlockLayer.EnableStatus.ENABLED;
   }

   /**
    * Implementation of the trigger search algorithm in the stack of triggers. Each layer of the
    * stack is inspected for <code>det</code> type of trigger associated with <code>buff</code>.
    * If found it is returned, otherwise the search continue with the next level from the stack.
    * If no such trigger can be found, <code>null</code> is returned.
    * This method takes into account if respective trigger event was disabled and, in this case,
    * a special singleton object is returned.
    *
    * @param   det
    *          The database trigger type.
    * @param   bufferCls
    *          The buffer that raised the trigger event.
    * @param   property
    *          In the case of ASSIGN trigger, the name of the property that fired the trigger
    *
    * @return  The top-level instance of the trigger data of the <code>det</code> type associated
    *          with <code>buff</code>.
    */
   private TriggerData peekTrigger(DatabaseEventType det,
                                   Class<? extends Buffer> bufferCls,
                                   String property)
   {
      for (TriggerBlockLayer blockLayer : stack)
      // the proper order is guaranteed by deque.iterator() method
      {
         final TriggerData triggerData = blockLayer.getTrigger(det, bufferCls, property);
         if (triggerData == null)
         {
            // no trigger found at this level, search deeper
            continue;
         }

         // if disabledTrigger was returned either a trigger was found or not, it does not
         // matter as long as nothing will be executed (disabledTrigger methods are empty and
         // also isOverride() return true so schema trigger is also disabled).
         // if a normal trigger is found it will be executed along with schema trigger
         return triggerData;
      }

      // search failed, no trigger of type det was defined for buff
      return null;
   }
}
